#include "console.h"
#include "luaconfig.h"
#include "portal.h"
#include "sdl.h"
#include "sprite.h"
#include "textrenderer.h"
#include "tileset.h"
#include "tiletype.h"
#include "window.h"
#include "world.h"

#include <SDL2/SDL.h>
#include <string>
#include <vector>
#include <map>
#include <iostream>

// For rand()
#include <cstdlib>
#include <ctime>

const unsigned short int MaxInventory = 6;

struct Player {
	unsigned int x;
	unsigned int y;
	unsigned int aspect;
	unsigned int select;
	class World * world;
	class TileType * inventory[MaxInventory];
};

void main_bis(std::string conf_filename);

int main(int argc, char * argv[]) {
	std::string conf_filename = argc > 1 ? argv[1] : "conf.lua";
	try {
		main_bis(conf_filename);
	} catch(std::string error) {
		std::cerr << "ERROR: " << error << std::endl;
	}
}

void main_bis(std::string conf_filename) {
	std::map<unsigned int, class TileType *> tiletypes;
	std::vector<class World *> worlds;
	struct Player player;

	// Init configuration.
	class LuaConfig * conf = new LuaConfig("conf.lua");
	std::string title_txt = "Countless Shards: Regrow the Worlds";
	unsigned short int escape = 0;
	unsigned short int timer = 0;
	unsigned short int timer_max = conf->get_int("growrate");
	unsigned short int game_age = 0;
	unsigned short int tile_size = conf->get_int("tile_size");
	bool victory = false;

	// Init SDL.
	class Sdl * sdl = new Sdl(
			conf->get_int("screen_w"),
			conf->get_int("screen_h"));

	// Load the tileset.
	class Tileset * tileset = new Tileset(tile_size, tile_size);
	for(std::string filename : conf->get_strings("tileset")) {
		tileset->append(sdl, filename);
	}

	// Load the font.
	class TextRenderer * font = new TextRenderer(
			sdl,
			conf->get_string("font_file"),
			conf->get_int("font_height"));

	// Create the console.
	class Console * console = new Console(
			font,
			conf->get_int("screen_w"),
			conf->get_int("console_height")*font->lineskip(),
			0, // x_shift
			conf->get_int("screen_h") // y_shift
				- conf->get_int("console_height")*font->lineskip()
			);

	// Create the window.
	class Window * window = new Window(
			conf->get_int("screen_w") / tile_size,
			(
				conf->get_int("screen_h")
				- (font->lineskip() + tile_size) // inventory height
				- conf->get_int("console_height")*font->lineskip()
			) / tile_size,
			tile_size,
			tile_size);
	unsigned int inventory_y =
				conf->get_int("screen_h")
				- (font->lineskip() + tile_size) // inventory height
				- conf->get_int("console_height")*font->lineskip();
	sdl->set_title(title_txt);
	if(conf->get_string("icon") != "") {
		sdl->set_icon(conf->get_string("icon"));
	}

	// Init RNG.
	srand(time(nullptr));

	// Display Intro

	std::vector<std::string> intro_txt;
	std::vector<class Sprite *> intro;
	intro_txt.push_back(title_txt);
	intro_txt.push_back(" ");
	intro_txt.push_back("* * *");
	intro_txt.push_back(" ");
	intro_txt.push_back("The Creator built a world, was unsatisfied and destroyed it.");
	intro_txt.push_back("You are a nature spirit of this shattered world, bound to heal and regrow it.");
	intro_txt.push_back("You can only walk on plants but you can harvest seeds and magically grow them.");
	intro_txt.push_back("You need to find a way to reach distant places and other world's shards.");
	intro_txt.push_back("Beware! There is other spirits, and they made a different choice than you.");
	intro_txt.push_back(" ");
	intro_txt.push_back("Use arrows to move,");
	intro_txt.push_back("space to harvest or plant");
	intro_txt.push_back("and tab to cycle through your inventory.");
	for(int i=0; i<intro_txt.size(); i++) {
		intro.push_back(font->render(sdl, intro_txt[i]));
	}
	while(not sdl->has_quit() and not sdl->any_keydown()) {
		for(int i=0; i<intro.size(); i++) {
			intro[i]->draw_center(sdl, sdl->get_width()/2, font->lineskip()*(i+2));
		}
		sdl->next_frame();
	}
	intro.clear();
	intro_txt.clear();

	// Populate Tiles

	tiletypes[' '] = new TileType(0, "void", false, 9999);
	tiletypes['@'] = new TileType(1, "player", false);
	tiletypes['P'] = new TileType(14, "portal", true);

	tiletypes['_'] = new TileType(3, "dirt", false, 0);
	tiletypes['s'] = new TileType(13, "sand", false, 100);
	tiletypes['o'] = new TileType(4, "rock", false, 200);

	tiletypes['?'] = new TileType(11, "black death spirit", false, 0);
	tiletypes['?']->flag_as_death_spirit();
	tiletypes['!'] = new TileType(12, "white death spirit", false, 150);
	tiletypes['!']->flag_as_death_spirit();
	tiletypes['$'] = new TileType(11, "black plague spirit", false, 5, TileType::Self, 3, Pattern::Grass);
	tiletypes['$']->flag_as_death_spirit();
	tiletypes['%'] = new TileType(12, "white plague spirit", false, 150, TileType::Self, 3, Pattern::Spores);
	tiletypes['%']->flag_as_death_spirit();

	tiletypes[','] = new TileType(15, "moss", true, 0);
	tiletypes['.'] = new TileType(2, "grass", true, 10, TileType::Self, 1, Pattern::Grass);
	tiletypes['|'] = new TileType(5, "vertical vines", true, 110, TileType::Self, 3, Pattern::VineVertical);
	tiletypes['-'] = new TileType(6, "horizontal vines", true, 110, TileType::Self, 3, Pattern::VineHorizontal);
	tiletypes['"'] = new TileType(7, "tall grass", true, 210);
	tiletypes['f'] = new TileType(10, "flower bush", true, 211, tiletypes['"'], 4, Pattern::Spores);

	tiletypes['p'] = new TileType(8, "purple chessgrass", true, 11, nullptr, 2, Pattern::Grass);
	tiletypes['c'] = new TileType(9, "cyan chessgrass", true, 11, nullptr, 2, Pattern::Grass);
	tiletypes['p']->set_output(tiletypes['c']);
	tiletypes['c']->set_output(tiletypes['p']);

	// Generate Worlds

	worlds.push_back(new World(70, 30, tiletypes['o']));	// #0: Tuto pickup.
	worlds.push_back(new World(70, 40, tiletypes['s']));	// #1: Tuto vines.
	worlds.push_back(new World(40, 40, tiletypes['o']));	// #2: Tuto carry.
	worlds.push_back(new World(70, 50, tiletypes['o']));	// #3: Advanced vines.
	worlds.push_back(new World(50, 50, tiletypes['o']));	// #4: Bonus chessgrass.
	worlds.push_back(new World(30, 30, tiletypes['o']));	// #5: Boss Arena.

	// #0: Tuto pickup.
	worlds[0]->fill_circle(tiletypes['_'], 15, 15, 10);		// Dirt circle...
	worlds[0]->fill_circle(tiletypes['.'], 15, 15, 5); 		// ... with grass circle inside.

	worlds[0]->fill_circle(tiletypes['"'], 30, 15, 6); 		// Then a tall grass "wall"...
	worlds[0]->fill_circle(tiletypes['_'], 45, 15, 10);		// ... and a dirt circle after it...
	worlds[0]->set(45, 15, tiletypes['?']);			   		// ... with a death spirit in it.

	worlds[0]->fill_circle(tiletypes['"'], 60, 15, 6); 		// In the end: a tallgrass circle...
	Portal::create_pair(worlds[0], 60, 15, worlds[1], 15, 20); // ... with a portal in it.

	// #1: Tuto vines.
	worlds[1]->fill_circle(tiletypes['.'], 20, 20, 10);		// In a big grass circle...
	worlds[1]->fill_circle(tiletypes['"'], 15, 20, 3);		// ... some tallgrass around the portal...
	worlds[1]->fill_circle(tiletypes['"'], 25, 20, 3);		// ... and some other further...
	worlds[1]->fill_square(tiletypes['-'], 20, 19, 20, 21); // ... and some vines in the middle.

	worlds[1]->fill_circle(tiletypes['_'], 40, 20, 5);		// After the sand, there is a circle of dirt...
	worlds[1]->fill_circle(tiletypes['"'], 45, 20, 2);		// ... with some tallgrass...
	Portal::create_pair(worlds[1], 45, 20, worlds[2], 10, 10);// ... around a portal...
	worlds[1]->set(37, 20, tiletypes['?']);					// ... and a death spirit in front of it.

	worlds[1]->fill_square(tiletypes['o'], 50, 0, 70, 40);	// The left side is filled with rocks...
	worlds[1]->fill_circle(tiletypes['s'], 60, 20, 7);		// ... with a sand island...
	worlds[1]->set(60, 16, tiletypes['?']);
	worlds[1]->set(60, 20, tiletypes['?']);					// ... where there is 3 death spirits...
	worlds[1]->set(60, 24, tiletypes['?']);
	Portal::create_pair(worlds[1], 65, 20, worlds[3], 10, 10);// ... and a portal behind them.
	// Three sand lines allow to reach the spirits and portal with vines :
	worlds[1]->fill_square(tiletypes['s'], 50, 16, 59, 16);
	worlds[1]->fill_square(tiletypes['s'], 50, 20, 59, 20);
	worlds[1]->fill_square(tiletypes['s'], 50, 24, 59, 24);
	// There is also two unreachable patches of color grass, one north and one south...
	worlds[1]->fill_circle(tiletypes['c'], 20, 5, 3);
	worlds[1]->fill_circle(tiletypes['p'], 20, 35, 3);
	// ...with a portal in each leading to the same shard (#4: Bonus chessgrass).
	Portal::create_pair(worlds[1], 20,  5, worlds[4], 25, 10);
	Portal::create_pair(worlds[1], 20, 35, worlds[4], 25, 40);

	// #2: Tuto carry.
	worlds[2]->fill_circle(tiletypes['_'], 10, 20, 6);
	worlds[2]->fill_circle(tiletypes['s'], 20, 30, 6);
	worlds[2]->fill_circle(tiletypes['s'], 30, 20, 6);

	worlds[2]->fill_circle(tiletypes['_'], 10, 10, 8);
	worlds[2]->fill_circle(tiletypes['"'], 10, 10, 2);
	worlds[2]->fill_circle(tiletypes['_'], 30, 10, 8);
	worlds[2]->fill_circle(tiletypes['"'], 30, 10, 2);
	worlds[2]->fill_circle(tiletypes['_'], 10, 30, 8);
	worlds[2]->fill_circle(tiletypes['"'], 10, 30, 2);
	worlds[2]->fill_circle(tiletypes['_'], 30, 30, 8);
	worlds[2]->fill_circle(tiletypes['"'], 30, 30, 2);
	worlds[2]->set(30, 10, tiletypes['f']); // Flower bush at the end...
	worlds[2]->set(26, 10, tiletypes['?']);
	worlds[2]->set(34, 10, tiletypes['?']); // ... surrounded by death spirits.
	worlds[2]->set(30, 6,  tiletypes['!']); // The last can only be killed by flowerbush.
	Portal::create_pair(worlds[2], 30, 30, worlds[3], 30, 40);

	// #3: Advanced vines.
	worlds[3]->fill_circle(tiletypes['s'], 20, 10, 5);	// Some sand bridges.
	worlds[3]->fill_circle(tiletypes['s'], 20, 40, 5);
	worlds[3]->fill_circle(tiletypes['s'], 30, 25, 9);
	worlds[3]->fill_circle(tiletypes['s'], 40, 25, 7);

	worlds[3]->fill_circle(tiletypes['.'], 10, 10, 7);	// Top left has the portal to #1: Tuto vines...
	worlds[3]->fill_circle(tiletypes['-'], 20, 10, 1);	// ... and a vine bridge eastward...
	worlds[3]->fill_circle(tiletypes['"'], 10, 10, 2);	// ... with some tallgras to stop it westward.

	worlds[3]->fill_circle(tiletypes['_'], 30, 10, 8);	// Top right is dirt...
	worlds[3]->fill_circle(tiletypes['|'], 38, 10, 2);	// ... with some vertical vines...
	worlds[3]->set(30, 5,  tiletypes['?']);				// ... and a little death spirit.

	worlds[3]->fill_circle(tiletypes['_'], 10, 40, 7);	// Bottom left is dirt...
	worlds[3]->set(10, 40, tiletypes['$']);				// ... with a plague spirit.

	worlds[3]->fill_circle(tiletypes['_'], 30, 40, 7);	// Bottom right is dirt...
	worlds[3]->fill_circle(tiletypes['"'], 30, 40, 2);	// ... with some tallgrass arount the portal to #2: Carry tuto.

	worlds[3]->fill_circle(tiletypes['s'], 50, 25, 10);	// Rightmost is a large sand circle...
	worlds[3]->set(50, 25, tiletypes['!']);				// ... with a big death spirit...
	worlds[3]->set(44, 27, tiletypes['?']);
	worlds[3]->set(47, 19, tiletypes['?']);				// ... some other minions...
	worlds[3]->set(54, 29, tiletypes['?']);
	worlds[3]->fill_circle(tiletypes['o'], 60, 25, 3);	// ... and a portal enceased in rock.
	Portal::create_pair(worlds[3], 60, 25, worlds[5], 15, 15);

	// #4: Bonus chessgrass.
	worlds[4]->fill_circle(tiletypes['.'], 25, 25, 23);		// This shard is mostly covered by grass.
	worlds[4]->fill_square(tiletypes['"'], 20, 5, 30, 15);
	worlds[4]->fill_square(tiletypes['c'], 21, 6, 29, 14);	// North: cyan grass squared in tallgrass.
	worlds[4]->fill_square(tiletypes['"'], 20, 35, 30, 45);
	worlds[4]->fill_square(tiletypes['p'], 21, 36, 29, 44);	// South: purple grass squared in tallgrass.

	worlds[4]->fill_square(tiletypes['"'], 5, 20, 15, 30);
	worlds[4]->fill_square(tiletypes['.'], 6, 21, 14, 29);	// West: chessgrass squared in tallgrass.
	worlds[4]->set(10, 25, tiletypes['p']);

	// #5: Boss Arena.
	worlds[5]->fill_circle(tiletypes['s'], 15, 15, 13);		// This shard is just a grand sand circle...
	worlds[5]->fill_circle(tiletypes['"'], 15, 15, 2);		// ... with just some tallgrass around the portal...
	worlds[5]->set(7, 18, tiletypes['o']);					// ... some scattered rocks...
	worlds[5]->set(14, 24, tiletypes['o']);
	worlds[5]->set(17, 8,  tiletypes['o']);
	worlds[5]->set(23, 19, tiletypes['o']);
	worlds[5]->set(25, 13, tiletypes['o']);
	worlds[5]->set(10, 7,  tiletypes['o']);
	worlds[5]->set(10, 12, tiletypes['!']);					// ... and a bunch of death spirits.
	worlds[5]->set(20, 14, tiletypes['!']);
	worlds[5]->set(12, 20, tiletypes['?']);
	worlds[5]->set(19, 21, tiletypes['?']);
	worlds[5]->set(21,  9, tiletypes['?']);
	worlds[5]->set(6, 14, tiletypes['?']);

	// Set the player in middle of the first world.

	player.aspect = 1;
	player.world = worlds[0];
	player.x = 15;
	player.y = 15;
	player.select = 0;
	for(int i=0; i<MaxInventory; i++) {
		player.inventory[i] = nullptr;
	}
	window->set_center(player.x, player.y);

	// Loop

	while(not sdl->has_quit() and escape < 2) {

		// Draw

		window->draw(sdl, player.world, tileset);
		window->draw(sdl, player.aspect, tileset, player.x, player.y); // Draw the player.
		for(int i=0; i<MaxInventory; i++) { // Draw the inventory.
			unsigned short int shift = i * tile_size * 1.5 + tile_size/2;
			if(i == player.select) {
				tiletypes['o']->get_aspect(tileset)->draw(sdl, shift, inventory_y - tile_size/2);
				tiletypes['o']->get_aspect(tileset)->draw(sdl, shift, inventory_y + tile_size/2);
			}
			if(player.inventory[i] != nullptr) {
				player.inventory[i]->get_aspect(tileset)->draw(sdl, shift, inventory_y);
			} else {
				tileset->get(0)->draw(sdl, shift, inventory_y);
			}
		}
		if(victory) {
			class Sprite * s = font->render(sdl, "VICTORY");
			s->draw_center(sdl, sdl->get_width()/2, 30);
		}
		console->draw(sdl);
		sdl->next_frame();

		// Process Input

		if(sdl->keydown(SDL_SCANCODE_ESCAPE)) {			// Escape for quitting.
			escape++;
			if(escape < 2) {
				console->add_line(sdl, font, "Are you sure to exit? (There is no save.) Press escape again.");
			}
		} else if(sdl->any_keydown()) {
			escape = 0;
		}

		if(sdl->keydown(SDL_SCANCODE_TAB)) {			// Tab for cycling inventory.
			player.select = (player.select + 1) % MaxInventory;
		} else if(sdl->keydown(SDL_SCANCODE_SPACE)) {	// Space for harvesting/planting.
			if(player.inventory[player.select] == nullptr) {
				player.inventory[player.select] = player.world->get(player.x, player.y);
				console->add_line(sdl, font, "You harvest seeds: "
					+ player.inventory[player.select]->get_name());
			} else {
				unsigned int x = player.x;
				unsigned int y = player.y;
				class TileType * t = player.inventory[player.select];
				player.world->set(x, y, t);
				if(t->get_pattern() == Pattern::Grass) {
					if(x > 0)
						t->grow(player.world, x-1, y);
					if(x < player.world->get_width()-1)
						t->grow(player.world, x+1, y);
					if(y > 0)
						t->grow(player.world, x, y-1);
					if(y < player.world->get_height()-1)
						t->grow(player.world, x, y+1);
				}
				// console->add_line(sdl, font, "You plant seeds: " + t->get_name());
				player.inventory[player.select] = nullptr;
			}
		} else if(										// Arrows for moving.
			sdl->keydown(SDL_SCANCODE_UP) ||
			sdl->keydown(SDL_SCANCODE_DOWN) ||
			sdl->keydown(SDL_SCANCODE_LEFT) ||
			sdl->keydown(SDL_SCANCODE_RIGHT)
		) {
			unsigned int X = player.x;
			unsigned int Y = player.y;
			if(sdl->keydown(SDL_SCANCODE_UP) && player.y > 0) {
				Y--;
			} else if(sdl->keydown(SDL_SCANCODE_DOWN) && player.y < player.world->get_height() -1) {
				Y++;
			} else if(sdl->keydown(SDL_SCANCODE_LEFT) && player.x > 0) {
				X--;
			} else if(sdl->keydown(SDL_SCANCODE_RIGHT) && player.x < player.world->get_width() -1) {
				X++;
			}
			class Portal * portal = player.world->get_portal(X,Y);
			if(portal) {
				player.world = portal->get_link()->get_world();
				player.x = portal->get_link()->get_x();
				player.y = portal->get_link()->get_y();
			} else if(player.world->get(X,Y)->is_passable()) {
				player.x = X;
				player.y = Y;
			}
			window->set_center(player.x, player.y);
		}

		// Grow

		if(timer >= timer_max) {
			timer = 0;
			game_age++;
			for(unsigned int w=0; w<worlds.size(); w++) {
				worlds[w]->grow(game_age, sdl, console, font);
			}
		} else {
			timer++;
		}

		// Check Victory
		if(not victory) {
			bool all_clear = true;
			for(int i=0; i<worlds.size(); i++) {
				if(worlds[i]->get_death_spirit_count() != World::NoDeathSpirit) {
					all_clear = false;
					break;
				}
			}
			if(all_clear) {
				victory = true;
				console->add_line(sdl, font, " * * *");
				console->add_line(sdl, font, "All reachable shards have been cleansed and regrown.");
				console->add_line(sdl, font, "You may now disapear in bliss, or continue playing with plants.");
			}
		}
	}

	worlds.clear();
	tiletypes.clear();
	delete(console);
	delete(font);
	delete(tileset);
	delete(window);
	delete(sdl);
}

