#include "world.h"

#include "tiletype.h"
#include "portal.h"
#include "sdl.h"
#include "tileset.h"
#include "window.h"
#include "sdl.h"
#include "console.h"
#include "textrenderer.h"

#include <algorithm> // min(), max()
#include <cmath> // sqrt()
#include <vector>
#include <set>

World::World(
	unsigned int width,
	unsigned int height,
	class TileType * base_type
) :
	width(width),
	height(height),
	tiles(std::vector<class TileType *>(width*height, 0)),
	death_spirits(World::UnknownDeathSpirit)
{
	this->fill_all(base_type);
}

World::~World() {
}

unsigned int World::get_width() {
	return(this->width);
}

unsigned int World::get_height() {
	return(this->height);
}

class TileType * World::get(unsigned int x, unsigned int y) {
	if(x < this->width && y < this->height) {
		return(this->tiles.at(y * this->width + x));
	} else {
		return(nullptr);
	}
}

void World::set(unsigned int x, unsigned int y, class TileType * val) {
	if(x < this->width && y < this->height) {
		this->tiles.at(y * this->width + x) = val;
	}
}

void World::add_portal(class Portal * p) {
	this->portals.insert(p);
}

void World::remove_portal(class Portal * p) {
	this->portals.erase(p);
}

class Portal * World::get_portal(unsigned int x, unsigned int y) {
	for(class Portal * p : this->get_portals()) {
		if(p->get_x() == x and p->get_y() == y) {
			return(p);
		}
	}
	return(nullptr);
}

std::vector<class Portal *> World::get_portals() {
	std::vector<class Portal *> v (this->portals.begin(), this->portals.end());
	return(v);
}

void World::fill_all(class TileType * val) {
	this->tiles.assign(this->width * this->height, val);
}

void World::fill_square(
	TileType * val,
	unsigned int lft,
	unsigned int top,
	unsigned int rgt,
	unsigned int bot
) {
	for(int x=std::min(lft,rgt); x<=std::max(lft,rgt); x++) {
		for(int y=std::min(top,bot); y<=std::max(top,bot); y++) {
			this->set(x,y,val);
		}
	}
}

void World::fill_circle(
	TileType * val,
	unsigned int x,
	unsigned int y,
	unsigned int radius
) {
	int R = radius;
	int X = x;
	int Y = y;
	int W = this->get_width()-1;
	int H = this->get_height()-1;
	for(int i=std::max(X-R,0); i<std::min(X+R,W); i++) {
		for(int j=std::max(Y-R,0); j<std::min(Y+R,H); j++) {
			if(std::sqrt(std::pow(X-i,2)+std::pow(Y-j,2)) <= R) {
				this->set(i,j,val);
			}
		}
	}
}

void World::grow(
	unsigned short int game_age,
	class Sdl * sdl,
	class Console * console,
	class TextRenderer * font
) {
	std::vector<struct GrowSave> saved;
	unsigned short int death_spirit_count = 0;

	// Detect growing tiles.
	for(unsigned int x=0; x < this->width; x++) {
		for(unsigned int y=0; y < this->height; y++) {
			class TileType * t = this->get(x,y);
			if(t->is_death_spirit()) {
				death_spirit_count++;
			} else if(t->get_pattern() != Pattern::Passive and game_age % t->get_timer() == 0) {
				struct GrowSave g = {x, y, t->get_output(), t->get_pattern()};
				saved.push_back(g);
			}
		}
	}

	// Update death spirits count.
	if(this->death_spirits == World::NoDeathSpirit) {
		// Nothing.
	} else if(this->death_spirits == World::UnknownDeathSpirit) {
		if(death_spirit_count == 0) {
			this->death_spirits = World::NoDeathSpirit;
		} else {
			this->death_spirits = death_spirit_count;
		}
	} else {
		if(death_spirit_count < this->death_spirits) {
			console->add_line(sdl, font, "A death spirit has been overwhelmed by growth.");
			this->death_spirits = death_spirit_count;
		}
		if(death_spirit_count == 0) {
			console->add_line(sdl, font, "A shard world has been cleansed from death spirits.");
			this->death_spirits = World::NoDeathSpirit;
		}
	}

	// Apply growth.
	for(int i=0; i<saved.size(); i++) {
		saved[i].output->grow(this, saved[i].x, saved[i].y, saved[i].pattern);
	}
}

int World::get_death_spirit_count() {
	return(this->death_spirits);
}

